# """
#     Import Syphe database in CSV format
# """
# __license__ = "GPL"
# __author__ = "Sylvain Meignier"
# __copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
# __maintainer__ = "Sylvain Meignier"
# __email__ = "pyDendron@univ-lemans.fr"
# __status__ = "Production"

# import logging
# from pathlib import Path
# import datetime
# import pickle

# import numpy as np 
# import pandas as pd

# from pyDendron.dataset import Dataset
# from pyDendron.app_logger import logger
# from pyDendron.dataname import *

# from pyDendron.tools.location import add_geocode

# sylphe_col_sequences = {   
#     'NumSeq': IDX, #int
#     'libellé': KEYCODE, #string  

#     'espèce':SPECIES, #string
#     #'NatureSeq' : 'object.type', #string
#     #'Forme': 'object.shape', # string

#     'NumGeo': SITE_CODE, #string
#     #'DouteGeo' :'place.doubt', # bool

#     'origine': DATE_BEGIN, #float
#     'terme' : DATE_END, #float
#     #'termeOptimum' : DATE_END_OPTIMUM, #float
#     #'termeMaxi' : DATE_END_MAXIMUM, #float
#     'DateSupose' : SYNC, # bool

#     'NumTypseq' : CATEGORY, #string
#     'NumSousTypSeq' : SUBCATEGORY, #string

#     #'Aubier' : SAPWOOD, # bool
#     'PosAubier' : SAPWOOD, #float

#     'Moelle' : PITH, # bool
#     'PosMoellerEstime' : PITH_ESTIMATED, #float

#     'Cambium' : CAMBIUM, # bool
#     #'PosCambiumEstimeMaxi' : CAMBIUM_MAXIMUM, #float
#     #'PosCambiumEstimeOpti': CAMBIUM_OPTIMUM, #float
#     #'typeEstimation': CAMBIUM_METHOD, #float
#     'CambiumType' : CAMBIUM_SEASON, #string

#     'Ecorce' : BARK, # bool

#     'DateCréation': CREATION_DATE, # date ?, #string
#     #'Informateur' : 'creation.owner', # string
#     'NumPersonne' : PERS_ID, # string
#     'Note': COMMENTS # string
#     }

# sylphe_col_sequences_miss = [
#         DATA_LENGTH, DATA_VALUES, DATA_TYPE, DATA_WEIGHTS,
#         ]

# sylphe_col_components = {'NumMean' : IDX_PARENT, 'NumSeq' : IDX_CHILD, 'Pos' : OFFSET}
# #sylphe_col_components_miss = ['keycode.child', 'path.child', ]
# sylphe_col_indices = {'NumSeq' : IDX, 'Longueur': 'count', 'TypeInd' : 'key', 'Val' : 'values'}
# sylphe_col_indices_miss = []

# class Sylphe:
#     """
#     Class to import Sylphe database after conversion in CSV files with '@' as separator
#     """

#     def __init__(self, get_place=False, places='location.p', reverse_places='reverse_location.p'):
        
#         self.sequences_db = None
#         self.indices_db = None
#         self.components_db = None

#         self.get_place = get_place
#         self.places = {}
#         self.places_fn = places
#         self.reverse_places_fn = reverse_places
#         self.elevations = {}

#         self.sequences = None
#         self.components = None
    
            
#     def read_sequences(self) :
#         """
#         Read cvs sequences

#         Keyword Arguments:
#             sep -- fields separator (default: {'@'})
#             header -- head column values (default: {0})
#             usecols -- columns to keep (default: 
#             rename_columns -- how to rename columns 
#         Returns:
#             a Pandas DataFrame and a list sequences with trouble
#         """

#         def _places():
#             def load(name):
#                 if (name is not None) and Path(name).exists():
#                     with open(name , 'rb') as fic:
#                         dic = pickle.load(fic)
#                 else:
#                     dic = {}
#                 return dic
#             def load2(name):
#                 if (name is not None) and Path(name).exists():
#                     with open(name , 'rb') as fic:
#                         dic = pickle.load(fic)
#                 else:
#                     dic = [{}, {}]
#                 return dic
#             places = load(self.places_fn)
#             reverse_places, elevation = load2(self.reverse_places_fn)
            
#             places, reverse_places, elevation = add_geocode(sequences, places, reverse_places, elevation)
#             with open(self.places_fn , 'wb') as fic:
#                 pickle.dump(places, fic)
                
#             with open(self.reverse_places_fn , 'wb') as fic:
#                 pickle.dump([reverse_places, elevation], fic)

#         def _gregorian():
#             # set gregorian date to astronomique gregorian date
#             for key in [DATE_BEGIN, DATE_END]: #, DATE_END_OPTIMUM, DATE_END_MAXIMUM]:
#                 sequences[key] = sequences[key].astype('Int32')
#                 sequences.replace({key : {0: pd.NA}}, inplace=True) 
#                 sequences.loc[sequences[key] < 0, key] += 1

#         def _type():
#             for key in [SAPWOOD]: 
#                 sequences[key] = sequences[key].apply(lambda x: pd.NA if (x > 159394000) or (x == 0) or (x == 65535) else x)
#                 sequences[key] = sequences[key].astype('Int32')
#                 sequences.loc[sequences[key] > 0, key] -= 1

#             for key in [BARK, CAMBIUM, PITH]:
#                 sequences[key] = sequences[key] > 0
#                 sequences[key] = sequences[key].astype(bool)

#         sequences = self.sequences_db[sylphe_col_sequences.keys()]
#         sequences.rename(columns=sylphe_col_sequences, inplace=True)
#         sequences = sequences[sylphe_col_sequences.values()]

#         for col in sylphe_col_sequences_miss:
#             sequences[col] = None

#         sequences.replace({CATEGORY: {'MEAN': CHRONOLOGY, 'SET': SET, 'SEQ': TREE}}, inplace=True)

#         _gregorian()
#         _type()
#         if self.get_place: _places()
        
#         self.sequences = sequences.set_index(IDX)
        
#         self._check_sequences()
        
#         return self.sequences

#     def _string(self, dataframe, key):
#         dataframe[key] = dataframe[key].astype('string')

#     def _check_sequences(self):
#         # Check lastYear and firstYear consistency
#         year = datetime.datetime.now().year
#         date1 = self.sequences.loc[(self.sequences[DATE_END] - self.sequences[DATE_BEGIN]) < 0].index.tolist()
#         date2 = self.sequences.loc[(self.sequences[DATE_END] > year) | 
#                                 (self.sequences[DATE_BEGIN] > year)].index.tolist()

#         trouble = list()
#         if date1 is not None:
#             trouble = date1
#         if date2 is not None:
#             trouble += date2
#         if len(trouble) > 0:
#             for idx in trouble:
#                 row = self.sequences.loc[idx]
#                 logger.warning(f'inconsistent dates, to ckeck: {idx} {row[KEYCODE]} (begin: {row[DATE_BEGIN]}, end:{row[DATE_END]})')

#         return trouble

#     def read_indices(self):
#         """Read CSV Indices

#         Keyword Arguments:
#             sep -- fields separator (default: {'@'})
#             header -- head column values (default: {0})
#         Returns:
#             a indices object and a list sequences with trouble
#         """
#         def count_nan_in_vector(vector):
#             if vector is None:
#                 return 0
#             return np.isnan(vector).sum()
        
#         def _read_values():
#             values = []
#             weights = []
#             tmp = row['values'].replace(',', '.')
#             tmp = tmp.replace('\\r\\n', '')
#             for line in tmp.split(';')[:-1]:
#                 try :
#                     s = line.split()
#                     x = s[1]
#                     w = s[3]
#                     x = np.nan if x == 'ABS' else float(x)
#                     x = np.nan if (x == 999) or (x == 9999) else x
#                 except ValueError:
#                     x = np.nan
#                     w = 0
#                     logger.warning(f'error in values {idx} {self.sequences.loc[idx, KEYCODE]} | {line}')
#                 weights.append(w)
#                 values.append(x)
            
#             if len(values) == 0:
#                 return pd.NA, 0, pd.NA
            
#             return np.array(values), len(values), np.array(weights)
        
#         def _check_dates(length):
#             keycode, begin, end = self.sequences.loc[idx, [KEYCODE, DATE_BEGIN, DATE_END]]
#             c = row['count']
#             s = f'end-begin dates != values length for: {idx} {keycode} : len != (DateEnd - DateBegin + 1)  {length} {c} != ( {end} - {begin} + 1 = {end - begin + 1})',  
#             if not pd.isna(end) and not pd.isna(begin) and length != (end - begin + 1):
#                 trouble.append([idx, length, begin, end])
#                 logger.warning(s)

#         # -------
#         #fn = Path(self.path) / Path(self.indices_filename)
#         #indices = pd.read_csv(fn, sep=sep, header=header, usecols=sylphe_col_indices.keys())
#         indices = self.indices_db[sylphe_col_indices.keys()]
#         indices.rename(columns=sylphe_col_indices, inplace=True)

#         # Get raw indices
#         indices = indices.loc[(indices['key'] == 'NAT') & (indices[IDX].isin(self.sequences.index)), :]
        
#         trouble = []
#         to_drop = []
#         for idx, df in indices.groupby(IDX): 
#             row = df.iloc[0]
#             values, length, weights = _read_values()
#             self.sequences.at[idx, DATA_VALUES] = values
#             self.sequences.at[idx, DATA_LENGTH] = length
#             self.sequences.at[idx, DATA_TYPE] = RAW

#             if self.sequences.at[idx, CATEGORY] == CHRONOLOGY:
#                 self.sequences.at[idx, DATA_WEIGHTS] = weights
#             _check_dates(length)

#             if (length == count_nan_in_vector(values)):
#                 if self.sequences.at[idx, CATEGORY] == TREE:
#                     to_drop.append(idx)
#                 else:
#                     self.sequences.at[idx, DATA_VALUES] = pd.NA
#                     self.sequences.at[idx, DATA_LENGTH] = 0
#                     self.sequences.at[idx, CATEGORY] = SET
#                     self.sequences.at[idx, DATA_WEIGHTS] = pd.NA
        
#         if len(to_drop) > 0:
#             logger.warning(f'Drop empty or full NaN {DATA_VALUES} for samples:  {to_drop}')
#             self.sequences.drop(to_drop, inplace=True)
        
#         for k, t in sequences_dtype_dict.items():
#             if k not in self.sequences.columns:
#                 self.sequences[k] = ''
#             if t == 'string':
#                 self.sequences[k] = self.sequences[k].fillna('')
        
#         self.sequences[SPECIES] = self.sequences[SPECIES].str.upper()
#         mask = self.sequences[SPECIES].str.startswith('QU')
#         self.sequences.loc[mask, SPECIES] = 'QU'
        
#         self.sequences[SUBCATEGORY] = self.sequences[SUBCATEGORY].str.upper()
#         self.sequences[COMMENTS] = self.sequences[COMMENTS].str.replace(r'\n', '.')
#         self.sequences[COMMENTS] = self.sequences[COMMENTS].str.replace(r'\r', '')
#         self.sequences[INCONSISTENT] = False
        
#     def read_components(self):
#         """_summary_

#         Keyword Arguments:
#             sep -- fields separator (default: {'@'})
#             header -- head column values (default: {0})
#         Returns:
#             a Pandas DataFrame and 3 lists with trouble
#         """

#         # ----------
#         #filename = Path(self.path) / Path(self.components_filename)
#         #components = pd.read_csv(filename, sep=sep, header=header, usecols=sylphe_col_components.keys())
#         components = self.components_db[sylphe_col_components.keys()]
#         components.rename(columns=sylphe_col_components, inplace=True)

#         self._add_missing_components_col(components)
#         key = OFFSET
#         components[key] = components[key].apply(lambda x: pd.NA if (x > 159394000) or (x == 0) or (x == 65535) else x)
#         components[key] = components[key].astype('Int32')
#         components.loc[components[key] > 0, key] -= 1
        
#         components_parents = self._idx_parent_is_in_sequences(components)
#         self._idx_child_is_in_sequences(components, components_parents)
        
#         components = self._check_components(components)

#         self.components = components
#         return self.components

#     def _add_missing_components_col(self, components):
#         pass
#         #for col in sylphe_col_components_miss:
#         #    components[col] = None

#     def _idx_parent_is_in_sequences(self, components):
#         owner_before = set(components[IDX_PARENT].to_list())
#         components_parents = components.loc[components[IDX_PARENT].isin(self.sequences.index)]
#         owner_after = set(components[IDX_PARENT].to_list())
#         owner_removed = owner_before - owner_after
#         if len(owner_removed) > 0:
#             logger.warning(f'remove components without sequence:  {owner_removed}')
#         return components_parents
        
#     def _idx_child_is_in_sequences(self, components, components_parents):
#         idx_before = set(components[IDX_CHILD].to_list())
#         components = components_parents.loc[components[IDX_CHILD].isin(self.sequences.index)]
#         idx_after = set(components[IDX_CHILD].to_list())
#         idx_removed = idx_before - idx_after

#         if len(idx_removed) > 0:
#             logger.warning(f'remove components without sequence:  {idx_removed}')
            

#     def _check_components(self, components):
#         components_with_date_begin = components.merge(self.sequences[DATE_BEGIN], left_on=IDX_CHILD, right_on=IDX)
        
#         for idx_parent, data in components_with_date_begin.groupby(IDX_PARENT):
#             min_date = data[DATE_BEGIN].min()
#             min_pos = data[OFFSET].min()
#             if pd.isna(min_date) or pd.isna(min_pos):
#                 components_with_date_begin.loc[components_with_date_begin[IDX_PARENT] == idx_parent, DATE_BEGIN] = 0
#                 components_with_date_begin.loc[components_with_date_begin[IDX_PARENT] == idx_parent, OFFSET] = 0
#             else:
#                 components_with_date_begin.loc[components_with_date_begin[IDX_PARENT] == idx_parent, DATE_BEGIN] -= min_date
#                 components_with_date_begin.loc[components_with_date_begin[IDX_PARENT] == idx_parent, OFFSET] -= min_pos
        
#         mask = (components_with_date_begin[OFFSET] != components_with_date_begin[DATE_BEGIN])
#         components_with_date_begin.loc[mask, OFFSET] = components_with_date_begin.loc[mask, DATE_BEGIN]
#         return components_with_date_begin.drop(DATE_BEGIN, axis=1)
    
#     def set_project(self):
#         components = self.dataset.components
#         leaf = self.dataset.get_leafs()
#         for idx in leaf:
#             filter = components.index.get_level_values(IDX_CHILD) == idx
#             if np.sum(filter) == 1:
#                 idx_parent = components[filter].index.get_level_values(IDX_PARENT).to_list()[0]
#                 self.dataset.sequences.at[idx, PROJECT] = self.dataset.sequences.at[idx_parent, KEYCODE]
        
#     def to_dataset(self, root_keycode='Dataset',root_idx=ROOT, trash_keycode='Trash', 
#                    clipboard_keycode='Clipboard', workshop_keycode='Workshop'):
#         logger.info('read sequences')
#         self.read_sequences()
#         logger.info('read indices')
#         self.read_indices()
#         logger.info('read components')
#         self.read_components()
#         #print(self.sequences)
#         self.dataset = Dataset(self.sequences, self.components, save_auto=False)
#         self.set_project()
        
#         logger.info(f'to_dataset roots: {self.dataset.get_roots()}')
        
#         self.dataset.new_root(root_keycode, root_idx)
#         self.dataset.new_clipboard(clipboard_keycode)
#         self.dataset.new_trash(trash_keycode)
#         self.dataset.new_workshop(workshop_keycode)
        
#         return self.dataset
 
#     def db_access(self, dbaccess):
#         try:
#             import pyodbc
#             #print('import_access')
#             logger.info('import_access')
#             connection_string = (
#                 r'DRIVER={Microsoft Access Driver (*.mdb, *.accdb)};'
#                 rf'DBQ={dbaccess};'
#             )
#             conn = pyodbc.connect(connection_string)
#             logger.info('import_access connection ok')

#             lst = []
#             for table, df in zip(['Sequence', 'Component', 'Index'], [self.sequences_access, self.components_access, self.indices_access]):
#                 query = f"SELECT * FROM {table}"
#                 lst.append(pd.read_sql(query, conn))
#             return lst
#         except Exception as inst:
#             logger.error(f'plot : {inst}', exc_info=True)
#             return None
   
#     def read_file(self, filename):
#         self.sequences_db, self.indices_db, self.components_db = self.db_access(filename)
#         keycode_parent = Path(filename).stem

#         return self.to_dataset(root_keycode=keycode_parent)



"""
Chronology
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"


import logging
import numpy as np
import pandas as pd
from concurrent.futures import ProcessPoolExecutor

from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *

epsilon = np.finfo(np.float64).tiny
biweight_factor = 9
#biweight_factor = 5

def data2col(data, key=IDX, use_offset=True, key_offset=OFFSET):
    
    def length_(values):
        l = len(values) if values is not None else 0
        #logger.debug(f'{l}, {values}')
        return l
    
    data = data.copy()
    data[DATA_LENGTH] = data[DATA_VALUES].apply(lambda x: length_(x))
    #print('data2col : use_offset', use_offset)
    if use_offset:
        if data[key_offset].isna().any():
            raise ValueError(f'data2col: NA value(s) in {key_offset} column')
        data[key_offset] -= data[key_offset].min()
        end = (data[key_offset]+data[DATA_LENGTH]).max()
    else:
        end = data[DATA_LENGTH].max()

    d = {}
    for idx, row in data.iterrows():
        offset = row[key_offset] if use_offset else 0
        values = row[DATA_VALUES]
        if (values is not None) and (len(values) > 0):
            vec = np.full(end, np.nan)
            vec[offset:offset+len(values)] = values
            k = idx if key == IDX else row[key]
            d[k] = vec
        else:
            raise ValueError(f'data2col: {idx} / {row[KEYCODE]}, ring values is missing')
        
    return pd.DataFrame(d)


def chronology(data, date_as_offset=False, ring_type='raw', biweight=False, idx_chronology=None):
    """
        sequences: DataFrame of samples. `offset` column is need if offset_type is `offset`
    """

    #print('chronology', idx_chronology)
    if OFFSET not in data.columns:
        raise ValueError('chronology: no offset in sequences')
    
    if date_as_offset:
        data[OFFSET] = data[DATE_BEGIN]

    if data[OFFSET].isna().any():
        key = DATE_BEGIN if date_as_offset else OFFSET
        raise ValueError(f'chronology: one or more {key} contain NA values')
            
    if not (data[CATEGORY] == TREE).all():
        raise ValueError('chronology : data need to contain sample only')
        #data = data[data[CATEGORY] == TREE]
    
    if not (data[DATA_TYPE] == ring_type).all():
        raise ValueError("chronology: ring type don't match data")
        
    data_col = data2col(data)
    if biweight:
        #logger.info('chronology : biweight mean')
        median = data_col.median(skipna=True, axis=1)
        data_center = data_col.sub(median, axis=0)
        kind_std = data_center.abs().median(skipna=True, axis=1)
        data_center_reduce = data_center.div((biweight_factor * kind_std) + epsilon, axis=0)
        data_weights = (1 - (data_center_reduce ** 2)) ** 2
        data_weights[np.abs(data_center_reduce) >= 1] = 0
        #print(data_weights)
        weights = np.sum(data_weights, axis=1)
        means = np.sum(data_weights*data_col, axis=1) / weights
    else:
        means = data_col.mean(axis=1)
        weights = data_col.count(axis=1)
    
    offsets = [(idx, row[OFFSET]) for idx, row in data.iterrows()]
    return np.round(means, 3).to_numpy(), weights.to_numpy(), offsets, idx_chronology

def chronologies(data_dict, date_as_offset=False, ring_type='raw', biweight=False, num_threads=1):
    res = dict()
    if num_threads == 1:
        for idx, data in data_dict.items():
            means, weights, offsets, idx = chronology(data, date_as_offset=date_as_offset, ring_type=ring_type, biweight=biweight, idx_chronology=idx)
            res[idx] = (means, weights, offsets)
    else:
        #perror(f'chronologies multithreading: num_threads {num_threads}')
        with ProcessPoolExecutor(max_workers=num_threads) as executor:
            futures = []
            # Start an asynchronous task (future) for each index
            for idx, data in data_dict.items():
                future = executor.submit(chronology, data, date_as_offset=date_as_offset, ring_type=ring_type, biweight=biweight, idx_chronology=idx)
                futures.append(future)

            # Wait for all tasks to be completed
            for future in futures:
                means, weights, offsets, idx = future.result()
                res[idx] = (means, weights, offsets)
    return res





import numpy as np
import pandas as pd
from pathlib import Path
import logging

import param
import panel as pn

from pyDendron.app_logger import logger, catch_bokeh_log, perror 
from pyDendron.dataset import Dataset
from pyDendron.gui_panel.sidebar import ParamColumns, ParamDetrend, ParamChronology, ParamPackage, ParamColumnStats
from pyDendron.gui_panel.dataset_selector import DatasetSelector
from pyDendron.gui_panel.dataset_treeview import DatasetTreeview
from pyDendron.gui_panel.dataset_package import DatasetPackage
from pyDendron.gui_panel.dataset_package_builder import DatasetPackageBuilder
from pyDendron.gui_panel.dataset_package_editor import DatasetPackageEditor
from pyDendron.gui_panel.tools_panel import ToolsPanel
from pyDendron.gui_panel.tabulator import tabulator
from pyDendron.gui_panel.ploter_panel import PloterPanel
from pyDendron.gui_panel.debug_panel import DebugPanel
from pyDendron.gui_panel.crossdating_panel import CrossDatingPanel
from pyDendron.crossdating import CrossDating
from pyDendron.chronology import data2col, chronology
from pyDendron.ploter import Ploter
from pyDendron.tools.alignment import Alignment
from pyDendron.alien.io_besancon import IOBesancon
from pyDendron.alien.io_heidelberg import IOHeidelberg
from pyDendron.alien.io_rwl import IORWL
from pyDendron.alien.io_sylphe import IOSylphe
from pyDendron.alien.io_dendronIV import IODendronIV
from pyDendron.alien.io_tridas import IOTridas
#from pyDendron.alien.sylpheII import SylpheII
#from pyDendron.alien.rwl import RWL
from pyDendron.estimation import cambium_estimation
from pyDendron.detrend import detrend, slope

def get_git_version():
    import subprocess
    
    try:
        import importlib.metadata

        version = importlib.metadata.version('pyDendron')
        #perror(f"Version metadata: {version}")
        return version
    except ImportError:
        # Get the directory of the __init__.py file
        current_directory = Path(__file__).resolve().parent
        file_path = current_directory / 'version.txt'
        # Exécute la commande git pour obtenir la version (tag) la plus récente
        version = subprocess.check_output(['git', 'describe', '--tags']).strip().decode('utf-8')

        #perror(f"Version git: {version}")
        return version
    except subprocess.CalledProcessError:
            return "Number unable."
    
__version__ = get_git_version()



"""
Tabulator tools
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import numpy as np
import pandas as pd
import panel as pn
import os
from pathlib import Path

from bokeh.models.widgets.tables import (NumberFormatter, DateFormatter, 
                                         SelectEditor, NumberEditor)

from pyDendron.dataname import *
from pyDendron.app_logger import logger, perror

def unique_filename(filename):
    filename = Path(filename)  
    directory = filename.parent  
    base, extension = filename.stem, filename.suffix  
    unique_fn = filename
    n = 1

    while unique_fn.exists():
        unique_fn = directory / f"{base}({n}){extension}"
        n += 1

    return unique_fn

def get_download_folder():
    # Pour Windows
    if os.name == 'nt':
        download_folder = Path(os.getenv('USERPROFILE')) / 'Downloads'
    # Pour macOS et Linux
    else:
        download_folder = Path.home() / 'Downloads'
    
    return download_folder

def _cell_transform(data):
    if data is None:
        return data
    data_out = pd.DataFrame()
    for col, dtype in data.dtypes.to_dict().items():
        if col not in [ICON, ICON2, IDX, IDX_CHILD, IDX_PARENT]:                    
            if str(dtype).lower().startswith('int'):
                data_out[col] = data[col].fillna(np.nan).astype('float')
            elif str(dtype).lower().startswith('float'):
                data_out[col] = data[col].fillna(np.nan).astype('float')
            elif str(dtype).lower().startswith('bool'):
                data_out[col] = data[col].astype('string').fillna('unk').str.lower()
            elif str(dtype).lower().startswith('string'):
                data_out[col] = data[col].fillna('')
            else:            
                data_out[col] = data[col]
        else:            
            data_out[col] = data[col]
    return data_out
            
def _cell_text_align(dtype_dict):
    aligns = {} 
    for key, dtype in dtype_dict.items():
        aligns[key] = 'left' if (dtype == 'string') or (dtype == 'object') else 'center' 
    if (ICON in aligns) or (ICON2 in aligns):
        aligns[ICON] = 'center'
    return aligns

def _cell_formatters(dtype_dict):
    formatters = {} 
    for key, dtype in dtype_dict.items():
        if dtype == 'int': formatters[key] = NumberFormatter(format='0')
        if dtype == 'Int32': formatters[key] = NumberFormatter(format='0')
        if dtype == 'float32': formatters[key] = NumberFormatter(format='0.000')
        #if dtype == 'boolean': formatters[key] = StringFormatter(nan_format = '-') #BooleanFormatter() #{'type': 'tickCross', 'allowEmpty': True, 'tickElement': "<i class='fa fa-check'></i>",'crossElement':"<i class='fa fa-times'></i>"} #BooleanFormatter(icon='check-square')
        if dtype == 'datetime64[ns]': formatters[key] = DateFormatter()
    if (ICON in dtype_dict) or (ICON2 in dtype_dict):
        formatters[ICON] = {'type': 'html'}
    return formatters

def _header_filters(dtype_dict):
    filters = {}
    for key, dtype in dtype_dict.items():
        if key not in [ICON, ICON2]:                
            if dtype == 'string': filters[key] =  {'type': 'input', 'func': 'like', 'placeholder': 'Like..'}
            if dtype == 'boolean': filters[key] = {'type': 'list', 'valuesLookup': True}
            if dtype == 'Int32': filters[key] = {'type': 'number', 'func': '=='}
            if dtype == 'float32': filters[key] = {'type': 'number', 'func': '=='}
    return filters

def _header_filters_lookup(dtype_dict):
    filters = {}
    for key, dtype in dtype_dict.items():
        if dtype == 'string': filters[key] =  {'type': 'list', 'func': 'in', 'valuesLookup': True, 'sort': 'asc', 'multiselect': True}
        if dtype == 'boolean': filters[key] = {'type': 'list', 'valuesLookup': True}
        if dtype == 'Int32': filters[key] = {'type': 'number', 'func': '=='}
        if dtype == 'float32': filters[key] = {'type': 'number', 'func': '>='}
    return filters

def _cell_editors(dtype_dict, edit=False):
    if edit == False:
        editors = {x:None for x in dtype_dict.keys()}
    else:
        editors = {}
        for key, dtype in dtype_dict.items():
            editors[key] = None
            if dtype == 'string': editors[key] =  {'type': 'list', 'valuesLookup': True, 'autocomplete':True, 'freetext':True, 'allowEmpty':True, }
            if dtype == 'Int32': editors[key] = NumberEditor(step=1) 
            if dtype == 'Float32': editors[key] = NumberEditor() 
            if dtype == 'boolean': editors[key] = SelectEditor(options=['true', 'false', 'unk'])
            if dtype == 'datetime64[ns]': editors[key] = 'date'
        if CATEGORY in dtype_dict:
            editors[CATEGORY] = SelectEditor(options=[SET, CHRONOLOGY, TREE])
        for col in [ICON, ICON2, IDX, IDX_CHILD, IDX_MASTER]:
            if col in dtype_dict:
                editors[col] = None            

    return editors

def tabulator(data):    
    return pn.widgets.Tabulator(data.reset_index(),
        pagination='local',
        header_filters=True, 
        sizing_mode='stretch_width',
        ) 

def _hidden_columns(column_list=[ICON, KEYCODE, DATE_BEGIN, DATE_END, OFFSET], dtype_view=dtype_view):
    return list(set(dtype_view.keys()) - set(column_list)) 

def _get_selection(wtabulator) -> pd.DataFrame:
    """
    Returns the view of selectionned rows. 
    """
    if wtabulator.pagination == 'remote':
        selection = wtabulator.value.iloc[wtabulator.selection]
    else: # 'local'
        idxs = [x for k, x in wtabulator._index_mapping.items() if k in wtabulator.selection]
        selection = wtabulator._processed.loc[idxs,:]
    #perror('-'*20)
    #perror(f'Selection: sorters {wtabulator.sorters}, filters {wtabulator.filters},  pagination {wtabulator.pagination}')
    #perror(f'Selection: {selection.index}')
    #perror('-'*20)

    return selection
    
    # is_sortered = len(wtabulator.sorters) > 0
    # d = wtabulator._index_mapping
    # is_filtered = sum([ k == v for k, v in d.items()]) != len(d)
    # is_filtered2 = len(wtabulator.filters) > 0
    
    # perror('-'*20)
    # perror(f'Selection: sorters {wtabulator.sorters}')
    # perror(f'Selection: filters {wtabulator.filters}')
    # perror(f'Selection: is_sortered: {is_sortered}, is_filtered: {is_filtered}, is_filtered2: {is_filtered2} pagination {wtabulator.pagination}')
    # idxs = [x for k, x in wtabulator._index_mapping.items() if k in wtabulator.selection]
    # selection = wtabulator._processed.loc[idxs,:]
    # perror(f'Selection: sorted / filtered{selection.index} local')
    # selection = wtabulator.value.iloc[wtabulator.selection]        
    # perror(f'Selection: sorted / not filtered{selection.index} remote')
    # selection = wtabulator.selected_dataframe
    # perror(f'Selection: not sorted / ? filtered{selection.index}')
    # perror('-'*20)


    # selection = None
    # if is_sortered:
    #     if is_filtered: # sorted and filtered
    #         perror('Selection: sorted and filtered')
    #         idxs = [x for k, x in wtabulator._index_mapping.items() if k in wtabulator.selection]
    #         selection = wtabulator._processed.loc[idxs,:]
    #     else: # sorted
    #         perror('Selection: sorted and not filtered')
    #         selection = wtabulator.value.iloc[wtabulator.selection]        
    # else: # not sorted and (filtered or not filtered)
    #     perror('Selection: not sorted and ? filtered')
    #     selection = wtabulator.selected_dataframe
    #     #if is_filtered:
    #     #    logger.warning('Selection not sorted, check restults. if results are OK, remove this warning.')

    #return selection

VALUES_PER_LINE = 20

def array2html(v):
    l = len(v)
    nl = (l + 1) // VALUES_PER_LINE + 2
    tmp = np.array([0.0] * nl * VALUES_PER_LINE, dtype=object)
    tmp[0:l] = v
    tmp[tmp == 0] = pd.NA
    tmp[len(v)] = ';'
    c = list(range(0, nl * VALUES_PER_LINE, VALUES_PER_LINE))
    return pd.DataFrame(tmp.reshape(-1, VALUES_PER_LINE).T, columns=c).T.style.format(precision=2)




"""
Dataset tools
"""
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"
__license__ = "GPL"

import pandas as pd
from pathlib import Path
import panel as pn
import param
import copy
import warnings
import os
from panel.viewable import Viewer

from pyDendron.dataset import Dataset
from pyDendron.alien.io_besancon import IOBesancon
from pyDendron.alien.io_heidelberg import IOHeidelberg
from pyDendron.alien.io_rwl import IORWL
from pyDendron.alien.io_sylphe import IOSylphe
from pyDendron.alien.io_dendronIV import IODendronIV
from pyDendron.alien.io_tridas import IOTridas
from pyDendron.tools.alignment import Alignment
from pyDendron.dataname import *
from pyDendron.gui_panel.dataset_package import DatasetPackage
from pyDendron.gui_panel.tabulator import (_cell_text_align, _cell_editors, _header_filters,
                                           _cell_formatters, _hidden_columns, unique_filename)

from pyDendron.app_logger import logger, notification_stream, perror
from pyDendron.gui_panel.my_viewer import MyViewer

class ToolsPanel(MyViewer):
    """ Tools to manage dataset. """
    path = param.Foldername('./', doc='Output path of the saved package.')
    filters = param.List(['*.p', '*.xlsx', '*.json'], doc='glob filter')
    encoding = param.Selector(default='utf-8', objects=['utf-8', 'latin1'], doc='file char encoding')
    tolerance = param.Integer(default=5, bounds=(1, 10), step=1, doc='Number of equal values tolerated.')
        
    def __init__(self, dataset, parameters, dendron_info, cfg_path, tmp_path, **params):
        super().__init__(cfg_path, **params)
        self.dataset = dataset
        self.parameters = parameters
        self.tmp_path = tmp_path
        self.wcolumn_selector = parameters.column.columns

        self.dataset_package = DatasetPackage(dataset, parameters.column, parameters.package, 
                                                      parameters.detrend, parameters.chronology, name='')
        self.dendron_info = dendron_info
        self.path = self.dendron_info.dataset_path
        self.www = self.dendron_info.www
        self.import_dataset = None
        self.outliers = []
        
        self.bt_size = 90

        self._layout = pn.Tabs(
            ('Import', self._import()),
            ('Export', self._export()),
            ('Merge', self._merge()),
            ('Detect', self._validate()),
            ('Edit rings', self._edit_ring()),
            name=self.name,
             dynamic=True, styles={'font-size': '15px'})

    def get_sidebar(self, visible=True):   
        return pn.Card(pn.Param(self, show_name=False),                 
                margin=(5, 0), 
                sizing_mode='stretch_width', 
                title='Tools',
                collapsed=True, visible=visible)
  
    def get_options(self):
        """ Retrieve file options based on filters. """
        options = {}
        for flt in self.filters:
            for file in Path(self.dendron_info.dataset_path).glob(flt):
                options[f'\U0001F4E6 {str(file.name)}'] = file
        return options
    
    def _import(self):

        def on_import(event):               
            """ Handle importing of datasets. """  
            def get_input_io(fn):
                tmp = fn.lower()
                if tmp.endswith('.mdb'):
                    return IOSylphe(base_directory=self.tmp_path)
                elif tmp.endswith('.fh') or tmp.endswith('.fh.zip'):
                    return IOHeidelberg()
                elif tmp.endswith('.rwl') or tmp.endswith('.crn') or tmp.endswith('.crn.zip') or tmp.endswith('.rwl.zip'):
                    return IORWL()
                elif tmp.endswith('.bes') or tmp.endswith('.bes.zip'):
                    return IOBesancon()
                elif tmp.endswith('.dxml.zip'):
                    return IODendronIV(base_directory=self.tmp_path)
                elif tmp.endswith('.tridas.zip') or tmp.endswith('tridas.xml'):
                    return IOTridas(base_directory=self.tmp_path)
                elif tmp.endswith('.p') or tmp.endswith('.json') or tmp.endswith('.xlsx'):
                    return Dataset(cfg_tmp=self.tmp_path)
                raise ValueError(f'Unknown file format {fn}, must be: mdb, fh, fh.zip, rwl, rwl.zip, crn, crn.zip, bes, bes.zip, Dxml.zip, tridas.zip, p, json, xlsx.')
            
            self._layout.loading = True
            try:    
                for fn, buffer in file_input.value.items():
                    io_import = get_input_io(fn)
                    parent_keycode = Path(fn).stem
                    notifications = pn.config.notifications
                    pn.config.notifications = False
                    ds = io_import.read_buffer(parent_keycode, buffer, file_input.mime_type[fn])
                    if self.import_dataset is None:
                        self.import_dataset = ds  
                    else:
                        self.import_dataset.append(ds, notify=False)
                    pn.config.notifications = notifications
                
                if self.import_dataset is not None:
                    #perror("get_data")
                    wtabulator.value = self.import_dataset.sequences
                else:
                    #perror("empty data")
                    wtabulator.value = pd.DataFrame(columns=list(dtype_view.keys()))
            except Exception as inst:
                logger.error(f'on_import : {inst}', exc_info=True)
            finally:
                self._layout.loading = False
                                
        
        def on_save(event):
            """ Handle saving of the imported dataset. """
            if self.import_dataset is None:
                raise ValueError('The imported dataset is empty.')
            if dataset_name.value == '':
                raise ValueError('The dataset name is empty.')      
            fn = unique_filename(Path(self.dendron_info.dataset_path) / dataset_name.value)      
            self.import_dataset.dump(fn)
            #logger.info(f'Save done in dataset {fn}')

        def on_merge(event):
            """ Handle saving of the imported dataset. """
            if self.import_dataset is None:
                raise ValueError('The imported dataset is empty.')
            if self.dataset is None:
                raise ValueError('The dataset is empty.')            
            self.dataset.append(self.import_dataset)
            #logger.info('Merge done in current dataset')

        
        """ Create the import tab layout. """

        def on_clear_import(event):
            self.import_dataset = None
            file_input.value = {}
            wtabulator.value = pd.DataFrame(columns=list(dtype_view.keys()))

        def sync_columns(event):
            wtabulator.hidden_columns = _hidden_columns(self.wcolumn_selector.value)

        option = ['Besançon', 'Heidelberg', 'RWL']
        option += ['Sylphe', 'DendronIV', 'pyDendron']
        
        file_input = pn.widgets.FileDropper(sizing_mode='stretch_both', multiple=True, max_file_size='1000MB')
        #file_input.param.watch(on_file_input, ['value'], onlychanged=True)
        
        #file_format = pn.widgets.RadioBoxGroup(name='Format', options=option, inline=True, align=('end', 'end'))
        bt_import = pn.widgets.Button(name='Import', icon='import', button_type='primary', align=('end', 'end'), description='Import the selected file.')
        bt_import.on_click(on_import)
        bt_clear_import = pn.widgets.Button(name='Clear import', icon='trash', button_type='primary', align=('end', 'end'), description='clear imported data.')
        bt_clear_import.on_click(on_clear_import)
        wtabulator = pn.widgets.Tabulator(
                                    pd.DataFrame(columns=list(dtype_view.keys())),
                                    hidden_columns=_hidden_columns(self.wcolumn_selector.value), 
                                    text_align=_cell_text_align(dtype_view),
                                    editors=_cell_editors(dtype_view), 
                                    header_filters=_header_filters(dtype_view), 
                                    formatters=_cell_formatters(dtype_view),
                                    sizing_mode='stretch_width',
                                    frozen_columns=[ICON, KEYCODE], 
                                    selectable='checkbox',
                                    pagination='local',
                                    page_size= 1000, 
                                    max_height=600,
                                    min_height=300,
                                    height_policy='max',
                                    layout='fit_data_fill',
                                    margin=(0,0),
                                    )
        self.wcolumn_selector.param.watch(sync_columns, ['value'], onlychanged=True, queued=True)
        
        dataset_name = pn.widgets.TextInput(name='Dataset name', placeholder='dataset_import.p', description='Enter the dataset name.')
        bt_save = pn.widgets.Button(name='Create', icon='file-plus', button_type='primary', align=('start', 'end'), description='Save the imported dataset in a new dataset.')
        bt_save.on_click(on_save)
        bt_merge = pn.widgets.Button(name='Merge in current dataset', icon='arrow-merge', button_type='primary', align=('start', 'end'), description='Merge the imported dataset in the current dataset.')
        bt_merge.on_click(on_merge)
        return pn.Column(file_input,
                        pn.Row(bt_clear_import, bt_import ),
                        wtabulator,
                        pn.Row(dataset_name, bt_save),
                        bt_merge
                    )

    def _export(self):
        def on_save_dataset(event):
            if self.dataset.filename is None:
                logger.warning('Dataset is not loaded.')
                return
            fn = Path(self.dataset.filename)
            if file_export_format.value == 'Pickle':
                fn = fn.with_suffix('.p')
            elif file_export_format.value == 'JSON':
                fn = fn.with_suffix('.json')
            else:
                fn = fn.with_suffix('.xlsx')
            fn = unique_filename(fn)
            logger.info(f'Tools export, filename: {fn} (format {file_export_format.value})')
            self.dataset.dump(fn)
            
        def on_save_dataset_package(event):
            if self.dataset.filename is None:
                logger.warning('Dataset is not loaded.')
                return
            data = dataset_package.data
            if data is None:
                logger.warning('No package to export.')
                return
            package = dataset_package.wselection.value.replace(os.sep, '_')
            if file_export_format_package.value == 'Besançon':
                bes = IOBesancon()
                fn = Path(self.path) / f'{package}.bes'
                bes.write_package(self.dataset, data, fn)
                #logger.info(f'Tools export {package}, filename: {fn} (format {file_export_format_package.value})')
            elif file_export_format_package.value == 'Heidelberg':
                fh = IOHeidelberg()
                fn = Path(self.path) / f'{package}.fh'
                fh.write_package(self.dataset, data, fn)
                #logger.info(f'Tools export {package}, filename: {fn} (format {file_export_format_package.value})')
            elif file_export_format_package.value == 'RWL':
                rwl = IORWL()
                fn = Path(self.path) / f'{package}.rwl'
                rwl.write_package(self.dataset, data, fn)
                #logger.info(f'Tools export {package}, filename: {fn} (format {file_export_format_package.value})')
              
        file_export_format = pn.widgets.RadioBoxGroup(name='file format', options=['Pickle', 'Excel', 'JSON'], inline=True, align=('start', 'end'))
        bt_save_dataset = pn.widgets.Button(name='Export', icon='file-arrow-right', button_type='primary', align=('start', 'end'), width=self.bt_size, description='Export the current dataset.')
        bt_save_dataset.on_click(on_save_dataset)
        
        file_export_format_package = pn.widgets.RadioBoxGroup(name='file format', options=['Heidelberg', 'Besançon'], inline=True, align=('start', 'end'))
        bt_save_dataset_package = pn.widgets.Button(name='Export', icon='file-arrow-right', button_type='primary', align=('start', 'end'), width=self.bt_size, description='Export the selected dataset.')
        bt_save_dataset_package.on_click(on_save_dataset_package)
        
        dataset_package = self.dataset_package
        return pn.Column(
                    pn.pane.Markdown('### Export current dataset'),
                    file_export_format,
                    bt_save_dataset,
                    pn.pane.Markdown('### Export current package'),
                    dataset_package,
                    file_export_format_package,
                    bt_save_dataset_package,
        )
  
    def _merge(self):
        """ Create the dataset merge tab layout. """
        
        def load(dataset_src_name1, dataset_src_name2):
            dataset_src = Dataset()
            dataset_src.load(dataset_src_name1.value)
            
            dataset_dest = Dataset()
            dataset_dest.load(dataset_src_name2.value)
            
            return  dataset_src, dataset_dest

        def on_merge(event):
            """ Handle merging of datasets. """
            try:
                self._layout.loading = True
                fn = unique_filename(Path(self.dendron_info.dataset_path) / dataset_dest_name.value)
                #perror(f'Merge {dataset_src_name1.value} and {dataset_src_name2.value} in {fn}')
                if dataset_src_name1.value == dataset_src_name2.value:
                    logger.warning('The source datasets are the same.')
                    return
                dataset_src1, dataset_src2 = load(dataset_src_name1, dataset_src_name2)
                dest = dataset_src1.clone()
                dest.filename = fn
                dest.append(dataset_src2)
                dest.dump()
                #logger.info(f'Merge done into {fn}')
            except Exception as inst:
                logger.error(f'on_duplicate2 : {inst}', exc_info=True)
            finally:
                self._layout.loading = False

        options = self.get_options()
        
        dataset_src_name1 = pn.widgets.Select(name='1st dataset name', options=options, description='Select the first dataset to merge.')
        dataset_src_name2 = pn.widgets.Select(name='2rd dataset name', options=options, description='Select the second dataset to merge.')
        dataset_dest_name = pn.widgets.TextInput(value='dataset_import.p', name='Destination Dataset name', description='Enter the destination dataset name.')

        bt_merge = pn.widgets.Button(name='Merge', icon='arrow-merge', button_type='primary', align=('start', 'end'), width=self.bt_size, description='Merge the 2 datasets.')
        bt_merge.on_click(on_merge)

        return pn.Column(
                        pn.Row(dataset_src_name1, dataset_src_name2), 
                        pn.Row(dataset_dest_name, bt_merge)
                    )
        
    def _validate(self):

        parent_orphans = []
        child_orphans = []
        seq_orphans = []
                
        def on_duplicate(event):
            try:
                self._layout.loading = True
                counts = self.dataset.sequences[KEYCODE].value_counts()
                lst = list(counts[counts > 1].to_dict().keys())
                wduplicate.value =  '\n'.join(lst) if len(lst) > 0 else 'No duplicate'
            except Exception as inst:
                logger.error(f'on_duplicate2 : {inst}', exc_info=True)
            finally:
                self._layout.loading = False

        def on_duplicate2(event):
            try:
                self._layout.loading = True
                data = dataset_package.data
                if data is None:
                    logger.warning('No package.')
                    return
                res = alignment.ndiff_sequences(data, n=self.tolerance)
                txt = ''
                for idx_i, idx_j, cost , insertions, deletions, substitutions, n, err in res:
                    #idx_i, idx_j, cost , insertions, deletions, substitutions, n, ch]
                    txt += f'{err} (ins: {insertions}, del:{deletions}, sub:{substitutions})\n'
                wduplicate2.value = txt
                progress.value = 100
            except Exception as inst:
                logger.error(f'on_duplicate2 : {inst}', exc_info=True)
            finally:
                self._layout.loading = False
        
        def on_progress(event):
            progress.value = alignment.rate
            
        def on_outliers(event):
            data = dataset_package.dt_data
            txt = ''
            self.outliers.clear()
            for _, row in data.iterrows():
                vec = row[DATA_VALUES]
                idx = row[IDX_CHILD]
                med, std = np.nanmedian(vec), np.nanstd(vec)
                cut_off = std * 5 
                lower, upper = med - cut_off, med + cut_off
                keycode = row[KEYCODE]
                for i, x in enumerate(vec):
                    if (x < lower) or (x > upper) or (abs(x) <= 1e-7):
                        txt += f'Outlier {x} in "{keycode}" at index {i} (lower={round(lower, 3)}, median={round(med, 3)}, upper={round(upper, 3)})\n'
                        self.outliers.append(({IDX: idx, KEYCODE: keycode, 'Index':i, 'Value': x}))
            woutliers.value = txt

        def on_set_nan(event):
            for d in self.outliers:
                self.dataset.sequences.at[d[IDX], DATA_VALUES][d['Index']] = np.nan
            self.dataset.notify_changes('outliers')
            
        def on_orphans(event):
            nonlocal parent_orphans, child_orphans, seq_orphans
            try:        
                self._layout.loading = True
                parent_orphans.clear()
                child_orphans.clear()
                seq_orphans.clear()
                parent_orphans += self.dataset.get_orphans_components(level=IDX_PARENT)
                child_orphans += self.dataset.get_orphans_components(level=IDX_CHILD)
                seq_orphans += self.dataset.get_orphans_sequences()
                worphans.value = f'Components {IDX_PARENT}:\n{parent_orphans}\n Components {IDX_CHILD}:\n{child_orphans}\n Sequences:\n{seq_orphans}'
            except Exception as inst:
                logger.error(f'on_orphans : {inst}', exc_info=True)
            finally:
                self._layout.loading = False
        
        def on_delete_orphans(event):
            try:
                self._layout.loading = True
                if len(parent_orphans) > 0:
                    self.dataset.drop_orphans_components(parent_orphans)
                if len(child_orphans) > 0:
                    self.dataset.drop_orphans_components(child_orphans)
                on_orphans(event)
            except Exception as inst:
                logger.error(f'on_delete_orphans : {inst}', exc_info=True)
            finally:
                self._layout.loading = False
        
        def on_check(event):
            try:
                self._layout.loading = True
                problems = []
                for idx, row in self.dataset.sequences.iterrows():
                    keycode = row[KEYCODE]
                    data = row[DATA_VALUES]
                    #perror(f'Check {keycode} {data}')
                    if isinstance(data, np.ndarray):
                        ldata = len(data)
                        if ldata != row[DATA_LENGTH]:
                            problems.append(f'In {keycode}, length {len(data)} different from {row[DATA_LENGTH]}.')
                        if pd.notna(row[SAPWOOD] ):
                            if row[SAPWOOD] >= ldata :
                                problems.append(f'In {keycode}, {SAPWOOD} value {row[SAPWOOD]} is greater than data lenght.')
                            elif row[SAPWOOD] < 0 :
                                problems.append(f'In {keycode}, {SAPWOOD} value {row[SAPWOOD]} is negative.')
                    if pd.notna(row[DATE_BEGIN]) and pd.notna(row[DATE_END]):
                        if row[DATE_BEGIN] > row[DATE_END]:
                            problems.append(f'In {keycode}, {DATE_BEGIN} {row[DATE_BEGIN]} is greater than {DATE_END}')
                        if isinstance(data, np.ndarray):
                            ldata = len(data)
                            if (row[DATE_END] - row[DATE_BEGIN] + 1) != ldata:
                                problems.append(f'In {keycode}, {DATE_END} - {DATE_BEGIN} + 1 = {row[DATE_END] - row[DATE_BEGIN] + 1} different from data length {ldata}.')                       
                wcheck.value = '\n'.join(problems) if len(problems) > 0 else 'No problem detected.'
            except Exception as inst:
                logger.error(f'on_check : {inst}', exc_info=True)
            finally:
                self._layout.loading = False
        
        dataset_package = self.dataset_package
        bt_duplicate = pn.widgets.Button(name=f'Detect duplicate {KEYCODE}', icon='bolt', button_type='primary', description='Detect duplicate in the current dataset.')
        bt_duplicate.on_click(on_duplicate)
        wduplicate = pn.widgets.TextAreaInput(name=f'Duplicate {KEYCODE}', value='', sizing_mode='stretch_width', description='List of duplicate keycodes.')

        alignment = Alignment()
        bt_duplicate2 = pn.widgets.Button(name=f'Detect duplicate {DATA_VALUES}', icon='bolt', button_type='primary', description=f'Detect duplicate {DATA_VALUES} in the current package. The tolerance is set to {self.tolerance}.')
        bt_duplicate2.on_click(on_duplicate2)
        progress = pn.indicators.Progress(name='Run', value=0, sizing_mode='stretch_width', disabled=True, bar_color='primary')
        alignment.param.watch(on_progress, ['rate'], onlychanged=True)
        wduplicate2 = pn.widgets.TextAreaInput(name=f'Duplicate {DATA_VALUES}', value='', sizing_mode='stretch_width', description=f'List of {KEYCODE} with duplicate values.')

        bt_outliers = pn.widgets.Button(name=f'Detect outliers {DATA_VALUES}', icon='bolt', button_type='primary', description=f'Detect outliers in the current package. Outliers are values outside 5 standard deviations from the median.')
        bt_outliers.on_click(on_outliers)
        woutliers = pn.widgets.TextAreaInput(name=f'Outliers {DATA_VALUES}', value='', sizing_mode='stretch_width', description='List of {KEYCODE} with outlier values.')
        bt_set_nan = pn.widgets.Button(name='Set value to NaN', icon='bolt', button_type='primary', description='Set the outlier values to NaN.')
        bt_set_nan.on_click(on_set_nan)

        bt_orphans = pn.widgets.Button(name='Detect', icon='bolt', button_type='primary', description='Detect orphan indexes in database.')
        bt_orphans.on_click(on_orphans)
        worphans = pn.widgets.TextAreaInput(name=f'{KEYCODE} orphans', value='', sizing_mode='stretch_width', description='List of detected orphans.')
        bt_delete_orphans = pn.widgets.Button(name='Delete', icon='delete', button_type='primary', description='Delete detected orphans.')
        bt_delete_orphans.on_click(on_delete_orphans)

        bt_check = pn.widgets.Button(name='Check', icon='bolt', button_type='primary', description=f'Check {DATA_LENGTH}, {SAPWOOD}, date.')
        wcheck = pn.widgets.TextAreaInput(name=f'Problems', value='', sizing_mode='stretch_width', description='List of detected problems.')
        bt_check.on_click(on_check)

        return pn.Tabs(
            (f'Duplicate {KEYCODE}', pn.Column(bt_duplicate, wduplicate)),
            (f'Duplicate {DATA_VALUES}', pn.Column(dataset_package, pn.Row(bt_duplicate2,  progress), wduplicate2)),
            (f'Outliers {DATA_VALUES}', pn.Column(dataset_package, bt_outliers, woutliers, bt_set_nan)),
            ('Orphans', pn.Column(bt_orphans, worphans, bt_delete_orphans)),
            ('Check data', pn.Column(bt_check, wcheck)),
            dynamic=True
        )

    def _edit_ring(self):
        def on_insert(event):
            if len(wtabulator.selection) > 0:
                df = wtabulator.value
                i = wtabulator.selection[0]
                with warnings.catch_warnings():
                    # TODO: pandas 2.1.0 has a FutureWarning for concatenating DataFrames with Null entries
                    warnings.filterwarnings("ignore", category=FutureWarning)
                    wtabulator.value = pd.concat([df.iloc[:i], pd.DataFrame({'value' : [pd.NA]}), df.iloc[i:]]).reset_index(drop=True)
        
        def on_append(event):
            with warnings.catch_warnings():
                # TODO: pandas 2.1.0 has a FutureWarning for concatenating DataFrames with Null entries
                warnings.filterwarnings("ignore", category=FutureWarning)
                wtabulator.value = pd.concat([wtabulator.value, pd.DataFrame({'value' : [pd.NA]})]).reset_index(drop=True)

        def on_delete(event):
            if len(wtabulator.selection) > 0:
                wtabulator.value = wtabulator.value.drop(wtabulator.selection).reset_index(drop=True)
                wtabulator.selection = []
        
        def on_save(event):
            try:
                self._layout.loading = True
                idx = select.value
                array = wtabulator.value['value'].to_numpy()
                old_len = dataset_package.dataset.sequences.at[idx, DATA_LENGTH]
                length = len(array)
                if length != old_len :
                    self.dataset_package.dataset.edit_sequence(idx, DATA_LENGTH, length, notify=False)                
                    self.dataset_package.dataset.edit_sequence(idx, DATA_LENGTH, dataset_package.dataset.sequences.at[idx, DATE_BEGIN] + length, notify=False)                
                ascendants = self.dataset_package.dataset.get_ascendants(idx, recursive=True)
                if len(ascendants) > 0:
                    self.dataset_package.dataset.edit_sequence(ascendants, INCONSISTENT, True, notify=False)
                self.dataset_package.dataset.edit_sequence(idx, DATA_VALUES, array)                
                wtabulator.value 
            except Exception as inst:
                logger.error(f'on_save : {inst}', exc_info=True)
            finally:
                self._layout.loading = False

        def sync_data(event):
            options = {}
            if dataset_package.dt_data is not None :
                for i, row in dataset_package.dt_data.iterrows():
                    options[row[KEYCODE]] = row[IDX_CHILD]
            select.options = options
            select.value = None

        def sync_tabulator(event):
            if select.value is not None:
                category = dataset_package.dataset.sequences.at[select.value, CATEGORY]
                if category == TREE:
                    data = dataset_package.dataset.sequences.loc[select.value, DATA_VALUES]
                    wtabulator.value = pd.DataFrame({'value' : data})
                else:
                    wtabulator.value = None
                    logger.warning(f'The selected serie is not a tree ring serie.')

        dataset_package = self.dataset_package
        dataset_package.wselection.param.watch(sync_data, ['value'], onlychanged=True)
        dataset_package.panel_tabulator.visible = False
        
        select = pn.widgets.Select(name=f'Select {KEYCODE}', options=[], description=f'Select the {KEYCODE} serie to edit.')
        select.param.watch(sync_tabulator, ['value'], onlychanged=True)
        
        bt_insert = pn.widgets.Button(name=f'Insert', icon='insert', button_type='primary', description='Insert a new value before the selected row.')
        bt_insert.on_click(on_insert)

        bt_append = pn.widgets.Button(name=f'Append', icon='append', button_type='primary', description='Append a new value after the last row.')
        bt_append.on_click(on_append)

        bt_delete = pn.widgets.Button(name=f'Delete', icon='delete', button_type='primary', description='Delete the selected rows.')
        bt_delete.on_click(on_delete)

        bt_save = pn.widgets.Button(name=f'Save', icon='save', button_type='primary', description='Save the modifications.')
        bt_save.on_click(on_save)

        wtabulator = pn.widgets.Tabulator(
                                    sizing_mode='stretch_width',
                                    max_height=600,
                                    min_height=300,
                                    height_policy='max',
                                    #selectable=True,
                                    selectable='checkbox',

                                    show_index = True
                                    )
        
        return pn.Column(
            dataset_package,
            select,
            wtabulator,
            pn.Row(bt_insert, bt_append, bt_delete, bt_save)
        )
     
    def __panel__(self):
        """ Return the panel layout."""
        return self._layout
    



        






import os
import getpass
import argparse
import sys

import panel as pn
from pathlib import Path
import json
#from bokeh.server.contexts import BokehSessionContext
import bokeh
from datetime import datetime
import platform
import shutil

from pyDendron.dataset import Dataset
from pyDendron.gui_panel.sidebar import ParamColumns, ParamDetrend, ParamChronology, ParamPackage, ParamColumnStats
from pyDendron.gui_panel.dataset_selector import DatasetSelector
from pyDendron.gui_panel.dataset_treeview import DatasetTreeview
from pyDendron.gui_panel.dataset_package import DatasetPackage
from pyDendron.gui_panel.dataset_package_builder import DatasetPackageBuilder
from pyDendron.gui_panel.dataset_package_editor import DatasetPackageEditor
from pyDendron.gui_panel.tools_panel import ToolsPanel
from pyDendron.gui_panel.ploter_panel import PloterPanel
from pyDendron.gui_panel.debug_panel import DebugPanel
from pyDendron.gui_panel.crossdating_panel import CrossDatingPanel
from pyDendron.app_logger import logger, perror, catch_bokeh_log, open_mail_client, LOG_FILENAME
from pyDendron.__init__ import __version__


def set_extension():
    """
    Set the required extensions for the pyDendron panel.

    This function enables the necessary extensions for the pyDendron panel to function properly.
    It enables the 'throttled', 'notifications', 'tabulator', and 'loading_spinner' extensions.
    It also sets the loading indicator to True and disables the console output.

    Parameters:
        None

    Returns:
        None
    """
    pn.extension('tabulator', 'filedropper', throttled=True, notifications=True, loading_spinner='dots')
    #pn.extension(defer_load=True, loading_indicator=True)
    
    pn.param.ParamMethod.loading_indicator = True

    pn.extension('terminal')
    pn.config.console_output = 'disable'

    pn.extension('tabulator')
    pn.extension(
        disconnect_notification="""Server Connected Closed <br /> <button class="btn btn-outline-light" onclick="location.reload();">Click To Reconnect</button> """
    )

class Configuration:
    """
    Class representing the parameters for pyDendron.
    """

    def __init__(self, app, cfg_path):
        self.app = app
        self.detrend = None
        self.chronology = None 
        self.column = None 
        self.column_stats = None
        self.package = None
        self.cfg_path = cfg_path
        os.makedirs(os.path.dirname(cfg_path), exist_ok=True)
        self.cfg_filename = cfg_path / Path('pyDendron.cfg.json')
        self.bt_save_parameters = pn.widgets.Button(name='Save parameters', icon="settings", button_type='default', align=('end', 'center'))
        self.bt_save_parameters.on_click(self.on_save_param)
        self.load()

    def get_widgets(self):
            """
            Returns the widget for saving parameters.

            Returns:
                bt_save_parameters (Widget): The widget for saving parameters.
            """
            return self.bt_save_parameters
    
    def save(self):
        """
        Save the parameters to a configuration file.

        This method serializes the parameters of various components and saves them to a JSON file.

        Returns:
            None
        """
        logger.info('save parameters')
        with open(self.cfg_filename, 'w') as f:
            data = {
                'detrend' : self.detrend.param.serialize_parameters(),
                'chronology' : self.chronology.param.serialize_parameters(),
                'column' : self.column.columns.value,
                'column_stat': self.column_stats.columns.value,
                'package' : self.package.param.serialize_parameters(),
            }
            json.dump(data, f)
        self.app.ploter.dump_cfg()
        self.app.crossdating.dump_cfg()
        self.app.dataset_selector.dump_cfg()
        self.app.tools.dump_cfg()
    
    def on_save_param(self, event):
        """
        Event handler for saving parameters.

        This method saves the current parameters and displays a notification
        indicating that the parameters have been saved.

        Args:
            event: The event object associated with the save action.

        Returns:
            None
        """
        self.save()
        logger.info('Parameters saved')
    
    def load(self):
        """
        Load the parameters from a configuration file.

        Args:
            cfg_filename (str): The path to the configuration file.

        Raises:
            Exception: If there is an error loading the configuration file.
        """
        try:
            if Path(self.cfg_filename).is_file():
                with open(self.cfg_filename, 'r') as f:
                    data = json.load(f)
                    self.detrend = ParamDetrend(**ParamDetrend.param.deserialize_parameters(data['detrend']))
                    self.chronology = ParamChronology(**ParamChronology.param.deserialize_parameters(data['chronology']))
                    self.column = ParamColumns(column_list=data['column'])
                    self.column_stats = ParamColumnStats(column_list=data['column_stat'])
                    self.package = ParamPackage(**ParamPackage.param.deserialize_parameters(data['package']))
            else:
                #print('no cfg')
                self.detrend = ParamDetrend()
                self.chronology = ParamChronology()
                self.column = ParamColumns()
                self.column_stats = ParamColumnStats()
                self.package = ParamPackage()
        except Exception as inst:
            logger.error(f'on_excel: {inst}', exc_info=True)
            logger.warning('ignore cfg files, version change.')
            self.detrend = ParamDetrend()
            self.chronology = ParamChronology()
            self.column = ParamColumns()
            self.column_stats = ParamColumnStats()
            self.package = ParamPackage()

class SidebarWidget:
    """
    A class representing a sidebar widget.

    Attributes:
        sidebar (object): The sidebar object.
        objects (dict): A dictionary to store the sidebar objects.
        hiddens (dict): A dictionary to store the paramters to hide depending of ative tab in main panel.
    """

    def __init__(self, app):
        # Add to sidebar
        self.sidebar = app.template.sidebar
        self.objects = {}
        self.hiddens = {}
        self.append_title('## Datasets')
        self.append(app.dataset_selector, 'dataset')
        self.append_title('## Parameters')
        self.append(app.configuration.column, 'param_column')
        self.append(app.configuration.package, 'param_package')
        self.append(app.configuration.detrend, 'param_detrend') 
        self.append(app.configuration.chronology, 'param_chronology')
        self.append(app.treeview, 'treeview')
        self.append(app.ploter, 'ploter')
        self.append(app.crossdating, 'crossdating')
        self.append(app.tools, 'tools')
        
        self.init_hiddens()
        self.active_parameters(app.treeview.name)
        
    def append(self, pane, name):
        """
        Append an parameter object to the sidebar.

        Args:
            pane (Viewer): The pane to be appended.
            name (str): The name of the object.
        """
        sidebar = pane.get_sidebar()
        self.objects[name] = sidebar
        self.sidebar.append(sidebar)

    def append_title(self, title):
        """
        Append a title to the sidebar.

        Args:
            title (str): The title to be appended.
            name (str): The name of the title.
        """
        self.sidebar.append(title)

    def init_hiddens(self):
        """
        Initialize the parameter objects to hide for a given the title of the tab.
        """
        self.hiddens['Treeview'] = ['ploter', 'crossdating', 'tools', 'param_detrend', 'param_package']
        self.hiddens['Package'] = ['treeview', 'ploter', 'crossdating', 'tools', 'param_detrend', 'param_chronology']
        self.hiddens['Ploter'] = ['treeview', 'crossdating', 'tools']
        self.hiddens['Crossdating'] = ['treeview', 'ploter', 'tools']
        #self.hiddens['Tools'] = ['treeview', 'ploter', 'crossdating', 'param_detrend', 'param_chronology']
        self.hiddens['Tools'] = ['treeview', 'ploter', 'crossdating', 'tools', 'param_colums', 'param_package', 
                                 'param_detrend', 'param_chronology']
        self.hiddens['Debug'] = ['treeview', 'ploter', 'crossdating', 'tools', 'param_colums', 'param_package', 
                                 'param_detrend', 'param_chronology']

    def active_parameters(self, name):
        """
        Activate the parameters based on the given name of the active tab.

        Args:
            name (str): The name to determine which parameters to activate.
        """
        lst = self.hiddens[name]
        for param_name, param in self.objects.items():
            param.visible = False if param_name in lst else True

class DendronApp:
    """
    The main application class for pyDendron.
    
    Args:
        dendron_info (object): An object containing information about pyDendron.
    """
    def __init__(self, www):
        self.www = www
        self.username = None
        self.dendron_path = None
        self.dataset_path = None
        self.cfg_path = None
        self.tmp_path = None

        self.set_directories()
        
        self.configuration = Configuration(self, self.cfg_path)
        self.dataset = Dataset(username=self.username)
        self.template = self.get_template()       
        filters = ['*.p', '*.json', '*.xlsx']
        self.dataset_selector = DatasetSelector(self.dataset, template=self.template, path=self.dataset_path, filters=filters, cfg_path=self.cfg_path )        
        self.treeview = DatasetTreeview(self.dataset, self.configuration, name='Treeview')
        self.tab_package = self.get_tab_package()
        self.ploter = PloterPanel(self.dataset, self.configuration, cfg_path=self.cfg_path, name='Ploter')
        self.crossdating = CrossDatingPanel(self.dataset, self.configuration, self.cfg_path, name='Crossdating')
        self.tools = ToolsPanel(self.dataset, self.configuration, self, tmp_path=self.tmp_path, cfg_path=self.cfg_path, filters=filters, name='Tools')
        self.debug = DebugPanel(self.dataset, self.configuration, self, name='Debug')
        
        self.tabs = pn.Tabs(self.treeview, self.tab_package, self.ploter, self.crossdating, self.tools, ('Debug', self.debug),
                dynamic=False, margin=0, styles={'padding' : '0px', 'font-size': '18px'}, sizing_mode='stretch_both',)
        self.tabs.param.watch(self.on_activate, ['active'], onlychanged=True)
        self.template.main.append(self.tabs)
        self.sidebar_widget = SidebarWidget(self)

    def set_directories(self):
        self.username = getpass.getuser() if pn.state.user is None else pn.state.user
        self.dendron_path = Path(os.path.expanduser("~")) / 'pyDendron'
        self.dataset_path = self.dendron_path / 'dataset' / Path(self.username)
        self.cfg_path = self.dendron_path / 'cfg' / Path(self.username)
        self.tmp_path = self.dendron_path / 'tmp' / Path(self.username)

        os.makedirs(self.dataset_path, exist_ok=True)
        os.makedirs(self.cfg_path, exist_ok=True) 
        os.makedirs(self.tmp_path, exist_ok=True) 
        
        dataset_test = self.dataset_path / 'dataset_test.p'
        dataset_test_src = os.path.join(os.path.dirname(__file__), 'data', 'dataset_test.p')
        if (not dataset_test.is_file()) and (Path(dataset_test_src).is_file()):
            shutil.copy(dataset_test_src, self.dataset_path)
        self.print_info()

    def print_info(self):
            logger.info(f'Mode: {self.www}')
            logger.info(f'username: {self.username}')
            logger.info(f'dendron_path: {self.dendron_path}')
            logger.info(f'dataset_path: {self.dataset_path}')
            logger.info(f'cfg_path: {self.cfg_path}')
            logger.info(f'tmp_path: {self.tmp_path}')


    def get_tab_package(self):
        """
        Get the tab package.
        
        Returns:
            pn.Tabs: The tab panel containing the package builder and package editor.
        """
        package_editor = DatasetPackageEditor(self.dataset, self.configuration, name='Editor')
        package_builder = DatasetPackageBuilder(self.dataset, self.configuration, self.treeview, name='Builder')
        tab_package = pn.Tabs(package_builder, package_editor, name='Package',
                                styles={'font-size': '16px'})
        return tab_package

    def bt_notification(self):
        """
        Create a button to clear notifications.
        
        Returns:
            pn.widgets.Button: The button to clear notifications.
        """
        def on_rm_notification(event):
            pn.state.notifications.clear()
            
        rm_notification = pn.widgets.Button(name='Clear notifications', icon="trash", button_type='default', align=('end', 'center'))
        rm_notification.on_click(on_rm_notification)
        return rm_notification
    
    def bt_gps(self):
        """
        Create a button to open the GPS site.
        
        Returns:
            pn.pane.Markdown: The button to open the GPS site.
        """
        # GPS 
        return pn.pane.Markdown("""<svg xmlns="http://www.w3.org/2000/svg" width="18" height="18" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="1" stroke-linecap="round" stroke-linejoin="round" class="icon icon-tabler icons-tabler-outline icon-tabler-map">
        <path stroke="none" d="M0 0h24v24H0z" fill="none" />
        <path d="M3 7l6 -3l6 3l6 -3v13l-6 3l-6 -3l-6 3v-13" />
        <path d="M9 4v13" />
        <path d="M15 7v13" />
        </svg>
        <a href="https://nominatim.openstreetmap.org" target="_blank">Open GPS site</a>""")

    def get_logout(self):
        return pn.pane.Markdown("""<svg  xmlns="http://www.w3.org/2000/svg"  width="18"  height="18"  viewBox="0 0 24 24"  fill="none"  stroke="currentColor"  stroke-width="1"  stroke-linecap="round"  stroke-linejoin="round"  class="icon icon-tabler icons-tabler-outline icon-tabler-logout">
            <path stroke="none" d="M0 0h24v24H0z" fill="none"/><path d="M14 8v-2a2 2 0 0 0 -2 -2h-7a2 2 0 0 0 -2 2v12a2 2 0 0 0 2 2h7a2 2 0 0 0 2 -2v-2" />
            <path d="M9 12h12l-3 -3" /><path d="M18 15l3 -3" /></svg>
            <a href="/logout" >Logout</a>""")    


    def bt_send_log(self):
        def on_send_log(event):
            
            panel_version = pn.__version__
            bokeh_version = bokeh.__version__
            date_today = datetime.now()
            os_info = platform.system() + ' ' + platform.release()
            user_name = pn.state.user #getpass.getuser()
            perror(f'User: {user_name}')
            
            agent = pn.state.headers["User-Agent"] if "User-Agent" in pn.state.headers else "Unknown"
            
            info_string = f"""
            pyDendron: {__version__}
            Date: {date_today}
            OS: {os_info}
            User: {user_name}
            Agent: {agent}
            Panel: {panel_version}
            Bokeh: {bokeh_version}"""

            info_string += '\n'+'-'*80 + '\n'
            with open(LOG_FILENAME, 'r') as file:
                info_string += file.read()
                
            open_mail_client(
                to='pyDendron@univ-lemans.fr',
                subject=f'pyDendron error {user_name}',
                body=info_string
            )
        
        bt_log = pn.widgets.Button(name='Send log', icon="send", button_type='default', align=('end', 'center'))
        bt_log.on_click(on_send_log)
        return bt_log

    def get_version(self):
        return pn.pane.Markdown(f"version: {__version__} user: {self.username}")

    
    def get_template(self):
        """
        Get the template for the application.
                
        Returns:
            pn.template.FastListTemplate: The template for the application.
        """
        logo = os.path.join(os.path.dirname(__file__), 'data', 'trees.svg')
        
        template = pn.template.FastListTemplate(title='pyDendron', 
                                                logo=logo,
                                                meta_author='LIUM, Le Mans Université',
                                                meta_keywords='Dendrochronology',
                                                main_layout=None,
                                                )
        template.header.objects = [self.get_version(), self.bt_notification(), self.configuration.get_widgets(), 
                                    self.bt_send_log(), self.bt_gps()]
        if self.www:
            template.header.objects.append( self.get_logout())
        template.modal.append(pn.Column())
        return template
            
    def on_activate(self, event):
        """
        Event handler for tab activation.
        
        Args:
            event (object): The event object.
        """
        self.sidebar_widget.active_parameters(self.tabs.objects[event.obj.active].name)
        
app = None
          
def cb_auto_save():
    """
    Automatically save the dataset.
    """
    if (app is not None) and app.dataset_selector.save_auto:
        logger.info('save dataset')
        app.dataset.dump()

def cb_disconnection(app):
    from pyDendron.app_logger import logger
    logger.info(f'On disconnection {pn.state.user} {app.username}.')
    if (app is not None) and app.dataset_selector.save_context:
        logger.info(f'Save context.')
        app.dataset.dump()
        app.configuration.on_save_param(None)
        #data = (app.dataset.filename, dendron_info, app.tabs.active, app.treeview.path)
        #pn.state.cache[f'context'] = data
        #logger.info(f'Save context: {data}')
        #raise KeyboardInterrupt

def cb_connection(app):
    pass
    # from pyDendron.app_logger import logger
    # import panel as pn
    # global app, dendron_info
    # try:
    #     self.dendron_info
    # except NameError:
    #     self.dendron_info = directory(args.www) 
    # try:
    #     app
    # except NameError:
    #     app = DendronApp(dendron_info)
    
    # logger.info(f'On connection')
    # if 'context' in pn.state.cache:
    #     logger.info('Load context')
    #     data = pn.state.cache['context']
    #     if data[1].username == app.dendron_info.username:
    #         app.dataset_selector.wselect.value = data[0]
    #         app.dataset_selector.on_load(None)
    #         app.treeview.path = data[3]
    #         app.treeview.show_data()
    #         app.tabs.active = data[2]
            

def interrupt(session_context):
    raise KeyboardInterrupt


app = None
#dendron_info = None

if __name__.startswith('bokeh_app'):
    parser = argparse.ArgumentParser(description="pydenron: A dendrochronology tool for tree-ring data analysis.")
    parser.add_argument('--www', action='store_true', help='A flag to enable www mode')
    args = parser.parse_args()
    www = args.www

    logger.info('Get paths')
    set_extension()
    logger.info('Create application ')
    app = DendronApp(www)
    logger.info('End of initialisation')
    app.template.servable()
    # Callback: autosave 
    logger.info('add_periodic_callback ')
    pn.state.add_periodic_callback(cb_auto_save, 1000*60*5) 
    catch_bokeh_log(cb_connection, cb_disconnection, app)


    



"""
Parametre classes for the sidebar
"""
__license__ = "GPL"
__author__ = "Sylvain Meignier"
__copyright__ = "Copyright 2023 Sylvain Meignier, Le Mans Université, LIUM (https://lium.univ-lemans.fr/)"

import os
import param
import panel as pn
from pyDendron.app_logger import logger, perror
from pyDendron.dataname import *

class ParamColumnStats:
    def __init__(self, column_list=[DATA_NAN, STAT_MEAN, STAT_MIN, STAT_MAX], title='Statistic Columns'):
        self.columns = pn.widgets.MultiChoice(name='Statistic Columns', value=column_list, options=list(stat_dtype_dict.keys()), 
                                     sizing_mode='stretch_width', search_option_limit=25, description='Select statistics to display in table.')
        self.title = title
    
    def get_columns(self):
        return  self.columns.value
    
    def get_widgets(self):
        return self.columns
    
    def get_sidebar(self):    
        return pn.Card(self.columns, title=self.title, sizing_mode='stretch_width', margin=(5, 0), collapsed=True)  

class ParamColumns:
    def __init__(self, column_list=[ICON, KEYCODE, DATE_BEGIN, DATE_END, OFFSET], title='Columns'):
        self.columns = pn.widgets.MultiChoice(name='Columns', value=column_list, options=list(dtype_view.keys()), 
                                     sizing_mode='stretch_width', search_option_limit=25, description='Select columns to display in table.')
        self.title = title

    def get_columns(self):
        return  self.columns.value
        
    def get_sidebar(self):    
        return pn.Card(self.columns, title=self.title, sizing_mode='stretch_width', margin=(5, 0), collapsed=True)  
    
class ParamChronology(param.Parameterized):
    num_threads = param.Integer(default=1, bounds=(1, os.cpu_count()), step=1, doc='Number of threads.')
    biweight_mean =  param.Boolean(False, doc='Biweight mean method for computing the chronology.')
    date_as_offset =  param.Boolean(True, doc='Dates are used as offsets for computing the chronology.')
    
    def get_sidebar(self):    
        return pn.Card(pn.Param(self, show_name=False), title='Chronology', sizing_mode='stretch_width', margin=(5, 0), collapsed=True)  

class ParamDetrend(param.Parameterized):
    num_threads = param.Integer(default=1, bounds=(1, 10), step=1, doc='Number of threads.')
    detrend = param.Selector(objects=ring_types, doc='Detrend method apply to the series.')
    window_size = param.Integer(default=5, bounds=(3, 15), step=2, doc='Size of the sliding window for detrending.')
    log = param.Boolean(True, doc='Perform a logarithm transformation at the end of detrending.')
            
    def __init__(self, **params):
        super(ParamDetrend,self).__init__(**params)
    
    def get_sidebar(self):
        return pn.Card(pn.Param(self, show_name=False), title='Detrend', sizing_mode='stretch_width', margin=(5, 0), collapsed=True)  

class ParamPackage(param.Parameterized):
    #show_data = param.Boolean(False, doc='Show data in selection view')
    cambium_estimation_method = param.Selector(default='Lambert', objects=['Lambert', 'log-log', 'user values'], doc='Method for cambium estimation.')
    lambert_parameters = param.Range(default=(12,23), bounds=(0, 60), step=1, doc='Lambert estimator range.')
    #pith_estimation = param.Boolean(False, doc='Draw pith estimation')
    slope_resolution = param.Integer(default=0, bounds=(0, 10), step=1, doc='Minimal GLK resolution')
            
    def __init__(self, **params):
        super(ParamPackage,self).__init__(**params)
    
    def get_sidebar(self):
        return pn.Card(pn.Param(self, show_name=False), title='Package', sizing_mode='stretch_width', margin=(5, 0), collapsed=True)  
